@@ -7,8 +7,9 @@ use v5.14;
 use utf8;
 
 package Parallel::WorkUnit;
+$Parallel::WorkUnit::VERSION = '1.001';
 # ABSTRACT: Provide easy-to-use forking with ability to pass back data
-$Parallel::WorkUnit::VERSION = '1.000';
+
 use strict;
 use warnings;
 use autodie;
@@ -35,7 +36,7 @@ sub BUILD {
     my $self = shift;
 
     $self->_subprocs( {} );
-    $SIG{CHLD} = \&_sigchld;
+    $SIG{CHLD} = 'IGNORE';
 }
 
 
@@ -68,7 +69,6 @@ sub async {
 
         try {
             my $result = $sub->();
-
             $self->_send_result( $pipe, $result );
         } catch($err) {
 
@@ -85,11 +85,7 @@ sub waitall {
     my ($self) = @_;
 
     my $sp = $self->_subprocs();
-    if ( !keys(%$sp) ) {
-
-        # We have no keys
-        return;
-    }
+    if ( !keys(%$sp) ) { return; }
 
     my $s = IO::Select->new();
     foreach ( keys(%$sp) ) { $s->add( $sp->{$_}{fh} ); }
@@ -98,11 +94,17 @@ sub waitall {
 
     foreach my $fh (@ready) {
         foreach my $child ( keys(%$sp) ) {
-            if ( $fh->fileno() == $sp->{$child}{fh}->fileno() ) {
-                $self->_read_result($child);
+            if ( defined($fh->fileno())) {
+                if ( $fh->fileno() == $sp->{$child}{fh}->fileno() ) {
+                    $self->_read_result($child);
+                }
             }
         }
     }
+   
+    # Just in case SIG{CHLD}="IGNORE" doesn't cause the children to
+    # be reaped. 
+    while ( ( my $child = waitpid( -1, WNOHANG ) ) > 0 ) { }
 
     # Tail recursion
     goto &waitall;
@@ -120,7 +122,13 @@ sub wait {
         return;
     }
 
-    $self->_read_result($pid);
+    my $result = $self->_read_result($pid);
+
+    # Just in case SIG{CHLD}="IGNORE" doesn't cause the children to
+    # be reaped. 
+    while ( ( my $child = waitpid( -1, WNOHANG ) ) > 0 ) { }
+
+    return $result;
 }
 
 sub _send_result {
@@ -141,12 +149,19 @@ sub _send {
     if ( $#_ != 3 ) { confess 'invalid call'; }
     my ( $self, $fh, $type, $data ) = @_;
 
-    print $fh $type, "\n";
+    $fh->write($type);
+    $fh->write("\n");
 
     my $msg = Storable::freeze( \$data );
 
-    print $fh length($msg), "\n";
+    $fh->write(length($msg));
+    $fh->write("\n");
+
+    binmode($fh, ':raw');
+
     $fh->write($msg);
+
+    $fh->close();
 }
 
 sub _read_result {
@@ -163,12 +178,24 @@ sub _read_result {
     my $size = <$fh>;
     chomp($size);
 
-    my $result;
-    $fh->read( $result, $size );
+    binmode($fh);
+
+    my $result = '';
+
+    my $part;
+    my $ret = 1;
+    while ( defined($ret) && ( length($result) < $size ) ) {
+        my $s = $size - length($result);
+
+        my $part = '';
+        $ret = $fh->read( $part, $s );
+        if ( defined($ret) ) { $result .= $part; }
+    }
 
     my $data = ${ Storable::thaw($result) };
 
     delete $self->_subprocs()->{$child};
+    $fh->close();
 
     if ( $type eq 'RESULT' ) {
         $cinfo->{callback}->($data);
@@ -177,15 +204,6 @@ sub _read_result {
     }
 }
 
-sub _sigchld {
-
-    # There are a bunch of subtle gotchas here.
-    # So we hopefully will rarely, if ever, see this fire
-    # unless the child has properly sent a result or an
-    # error.
-    while ( ( my $child = waitpid( -1, WNOHANG ) ) > 0 ) { }
-}
-
 __PACKAGE__->meta->make_immutable;
 
 1;
@@ -202,7 +220,7 @@ Parallel::WorkUnit - Provide easy-to-use forking with ability to pass back data
 
 =head1 VERSION
 
-version 1.000
+version 1.001
 
 =head1 SYNOPSIS
 
@@ -211,8 +229,7 @@ version 1.000
 
   $wu->waitall();
 
-There are many other Parallel::* applications in CPAN - it would be worth
-any developer's time to look through those and choose the best one.
+=head1 DESCRIPTION
 
 This is a very simple forking implementation of parallelism, with the
 ability to pass data back from the asyncronous child process in a
@@ -221,6 +238,9 @@ the information, serialized, back).  It was designed to be very simple
 for a developer to use, with the ability to pass reasonably large amounts
 of data back to the parent process.
 
+There are many other Parallel::* applications in CPAN - it would be worth
+any developer's time to look through those and choose the best one.
+
 =head1 METHODS
 
 =head2 new
@@ -257,9 +277,13 @@ to exit.  This method handles children that C<die()> or return
 a serializable data structure.  When all children return, this
 method will return.
 
-If a child dies, this method will C<die()> and propagate a
+If a child dies unexpectedly, this method will C<die()> and propagate a
 modified exception.
 
+This method has a side-effect of reaping any children that haven't yet
+been reaped.  If you aren't messing with C<$SIG{CHLD}>, you do not need
+to worry about this.
+
 =head2 wait($pid)
 
 This functions simiarly to C<waitall()>, but only waits for
@@ -271,6 +295,17 @@ executing, it simply returns.  Otherwise, it waits until the
 child process's work unit is complete and executes the callback
 routine, then returns.
 
+This method has a side-effect of reaping any children that haven't yet
+been reaped.  If you aren't messing with C<$SIG{CHLD}>, you do not need
+to worry about this.
+
+=head1 SIDE EFFECTS
+
+This module, when instantiated, sets the SIG{CHLD} handler to 'IGNORE'.
+This means that children will be automatically reaped on most systems.
+In addition, the C<wait()> and C<waitall()> methods will attempt to
+reap any outstanding children.
+
 =head1 AUTHOR
 
 Joel Maslak <jmaslak@antelope.net>
